/++
 + 書式モジュール
 +/
module jp.ku6.format;

import std.datetime;

/// format関数の%Aで使用する曜日の名前
immutable
string[7] DEFAULT_WEEKDAYS = ["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"];

/// format関数の%Bで使用する月の名前
immutable
string[12] DEFAULT_MONTHS = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"];

/++
 + 日時を書式に従って文字列に変換する
 +
 + 書式はUnixのDateコマンド形式
 + 具体的に変換するものは下記の通り
 + %Y 年(1970-)
 + %y 年(00-99)
 + %m 月(01-12)
 + %d 日(01-31)
 + %H 時(00-23)
 + %I 時(01-12)
 + %k 時(0-23)
 + %l 時(1-12)
 + %M 分(00-59)
 + %S 秒(00-59)
 + %D 日付(mm/dd/yy)
 + %T 24時間形式(hh:mm:ss)
 + %A 曜日(weekdaysを使用)
 + %B 月(monthsを使用)
 + %n 改行
 + %t タブ
 + %% %自体
 +
 + Params:
 +  fmt = 書式
 +  time = 時間
 +  weekdays = 曜日の文字列表現 省略した場合はデフォルトを使用
 +  months = 月の文字列表現 省略した場合はデフォルトを使用
 +
 + Return:
 +  引数の書式に従って展開された文字列
 +/
string format(string fmt, const(SysTime) time, string[7] weekdays = DEFAULT_WEEKDAYS, string[12] months = DEFAULT_MONTHS)
{
	ushort year;
	ubyte month;
	ubyte day;
	ubyte hour;
	ubyte ihour;
	ubyte minute;
	ubyte second;
	ubyte dayOfWeek;

	// まずは結果の長さを算出する
	uint resultLength;
	const(char)* iterator = fmt.ptr;
	const(char)* end = fmt.ptr + fmt.length;
	while (iterator < end)
	{
		if (*iterator == '%')
		{
			++ iterator;
			if (iterator >= end)
			{
				resultLength += 1;
				break;
			}
			else
			{
				switch(*iterator)
				{
				case '%':
				case 'n':
				case 't':
					resultLength += 1;
					break;
				case 'H':
					hour = time.hour;
					resultLength += 2;
					break;
				case 'I':
					ihour = time.hour;
					if (ihour > 12)
						ihour -= 12;
					else if (ihour == 0)
						ihour = 12;
					resultLength += 2;
					break;
				case 'M':
					minute = time.minute;
					resultLength += 2;
					break;
				case 'S':
					second = time.second;
					resultLength += 2;
					break;
				case 'm':
					month = time.month;
					resultLength += 2;
					break;
				case 'd':
					day = time.day;
					resultLength += 2;
					break;
				case 'y':
					year = time.year;
					resultLength += 2;
					break;
				case 'Y':
					year = time.year;
					resultLength += 4;
					break;
				case 'T':
					hour = time.hour;
					minute = time.minute;
					second = time.second;
					resultLength += 8;
					break;
				case 'D':
					year = time.year;
					month = time.month;
					day = time.day;
					resultLength += 8;
					break;
				case 'k':
					hour = time.hour;
					resultLength += hour > 9 ? 2 : 1;
					break;
				case 'l':
					ihour = time.hour;
					if (ihour > 12)
						ihour -= 12;
					else if (ihour == 0)
						ihour = 12;
					resultLength += ihour > 9 ? 2 : 1;
					break;
				case 'A':
					dayOfWeek = time.dayOfWeek;
					resultLength += weekdays[dayOfWeek].length;
					break;
				case 'B':
					month = time.month;
					resultLength += months[month - 1].length;
					break;
				default:
					resultLength += 2;
					break;
				}
			}
		}
		else
			resultLength += 1;

		++ iterator;
	}

	// 結果用バッファを確保
	char[] result = new char[resultLength];

	// 結果バッファに書き込む
	iterator = fmt.ptr;
	char* write = result.ptr;
	while (iterator < end)
	{
		if (*iterator == '%')
		{
			++ iterator;
			if (iterator >= end)
			{
				*(write++) = '%';
				break;
			}
			else
			{
				switch(*iterator)
				{
				case '%':
					*(write++) = '%';
					break;
				case 'n':
					*(write++) = '\n';
					break;
				case 't':
					*(write++) = '\t';
					break;
				case 'H':
					*(write++) = cast(char)(hour / 10 + 0x30);
					*(write++) = cast(char)(hour % 10 + 0x30);
					break;
				case 'I':
					*(write++) = cast(char)(ihour / 10 + 0x30);
					*(write++) = cast(char)(ihour % 10 + 0x30);
					break;
				case 'M':
					*(write++) = cast(char)(minute / 10 + 0x30);
					*(write++) = cast(char)(minute % 10 + 0x30);
					break;
				case 'S':
					*(write++) = cast(char)(second / 10 + 0x30);
					*(write++) = cast(char)(second % 10 + 0x30);
					break;
				case 'm':
					*(write++) = cast(char)(month / 10 + 0x30);
					*(write++) = cast(char)(month % 10 + 0x30);
					break;
				case 'd':
					*(write++) = cast(char)(day / 10 + 0x30);
					*(write++) = cast(char)(day % 10 + 0x30);
					break;
				case 'y':
					*(write++) = cast(char)((year % 100) / 10 + 0x30);
					*(write++) = cast(char)((year % 100) % 10 + 0x30);
					break;
				case 'Y':
					*(write++) = cast(char)(year / 1000 + 0x30);
					*(write++) = cast(char)((year % 1000) / 100 + 0x30);
					*(write++) = cast(char)(((year % 1000) % 100) / 10 + 0x30);
					*(write++) = cast(char)(((year % 1000) % 100) % 10 + 0x30);
					break;
				case 'T':
					*(write++) = cast(char)(hour / 10 + 0x30);
					*(write++) = cast(char)(hour % 10 + 0x30);
					*(write++) = ':';
					*(write++) = cast(char)(minute / 10 + 0x30);
					*(write++) = cast(char)(minute % 10 + 0x30);
					*(write++) = ':';
					*(write++) = cast(char)(second / 10 + 0x30);
					*(write++) = cast(char)(second % 10 + 0x30);
					break;
				case 'D':
					*(write++) = cast(char)(month / 10 + 0x30);
					*(write++) = cast(char)(month % 10 + 0x30);
					*(write++) = '/';
					*(write++) = cast(char)(day / 10 + 0x30);
					*(write++) = cast(char)(day % 10 + 0x30);
					*(write++) = '/';
					*(write++) = cast(char)((year%100) / 10 + 0x30);
					*(write++) = cast(char)((year%100) % 10 + 0x30);
					break;
				case 'k':
					if (hour > 9)
					{
						*(write++) = cast(char)(hour / 10 + 0x30);
						*(write++) = cast(char)(hour % 10 + 0x30);
					}
					else
					{
						*(write++) = cast(char)(hour + 0x30);
					}
					break;
				case 'l':
					if (ihour > 9)
					{
						*(write++) = cast(char)(ihour / 10 + 0x30);
						*(write++) = cast(char)(ihour % 10 + 0x30);
					}
					else
					{
						*(write++) = cast(char)(ihour + 0x30);
					}
					break;
				case 'A':
					string word = weekdays[dayOfWeek];
					size_t length = word.length;
					for (size_t i; i < length; ++i)
						*(write++) = word[i];
					break;
				case 'B':
					string word = months[month - 1];
					size_t length = word.length;
					for (size_t i; i < length; ++i)
						*(write++) = word[i];
					break;
				default:
					*(write++) = '%';
					*(write++) = *iterator;
					break;
				}
			}
		}
		else
			*(write++) = *iterator;

		++ iterator;
	}

	return cast(immutable)result;
}
unittest
{
	auto t = SysTime(DateTime(2011, 3, 2, 13, 4, 5));

	assert(std.string.cmp(format("%Y年", t), "2011年") == 0);
	assert(std.string.cmp(format("%y年", t), "11年") == 0);
	assert(std.string.cmp(format("%m月", t), "03月") == 0);
	assert(std.string.cmp(format("%d日", t), "02日") == 0);
	assert(std.string.cmp(format("%H時", t), "13時") == 0);
	assert(std.string.cmp(format("%I時", t), "01時") == 0);
	assert(std.string.cmp(format("%k時", t), "13時") == 0);
	assert(std.string.cmp(format("%l時", t), "1時") == 0);
	assert(std.string.cmp(format("%M分", t), "04分") == 0);
	assert(std.string.cmp(format("%S秒", t), "05秒") == 0);
	assert(std.string.cmp(format("=%D=", t), "=03/02/11=") == 0);
	assert(std.string.cmp(format("%%%T%%", t), "%13:04:05%") == 0);
	assert(std.string.cmp(format("%A%n", t), "Wednesday\n") == 0);
	assert(std.string.cmp(format("%B%t", t), "March\t") == 0);
	assert(std.string.cmp(format("%Y年%m月%d日 %H時%M分%S秒", t), "2011年03月02日 13時04分05秒") == 0);
}

